@@ -15,65 +15,64 @@ import os
 
 from s3transfer.manager import TransferManager
 
-from awscli.customizations.s3.utils import (
-    human_readable_size, MAX_UPLOAD_SIZE, find_bucket_key, relative_path,
-    create_warning, NonSeekableStream)
-from awscli.customizations.s3.transferconfig import \
-    create_transfer_config_from_runtime_config
-from awscli.customizations.s3.results import UploadResultSubscriber
-from awscli.customizations.s3.results import DownloadResultSubscriber
-from awscli.customizations.s3.results import CopyResultSubscriber
-from awscli.customizations.s3.results import UploadStreamResultSubscriber
-from awscli.customizations.s3.results import DownloadStreamResultSubscriber
-from awscli.customizations.s3.results import DeleteResultSubscriber
-from awscli.customizations.s3.results import QueuedResult
-from awscli.customizations.s3.results import SuccessResult
-from awscli.customizations.s3.results import FailureResult
-from awscli.customizations.s3.results import DryRunResult
-from awscli.customizations.s3.results import ResultRecorder
-from awscli.customizations.s3.results import ResultPrinter
-from awscli.customizations.s3.results import OnlyShowErrorsResultPrinter
-from awscli.customizations.s3.results import NoProgressResultPrinter
-from awscli.customizations.s3.results import ResultProcessor
-from awscli.customizations.s3.results import CommandResultRecorder
-from awscli.customizations.s3.utils import RequestParamsMapper
-from awscli.customizations.s3.utils import StdoutBytesWriter
-from awscli.customizations.s3.utils import ProvideSizeSubscriber
-from awscli.customizations.s3.utils import ProvideETagSubscriber
-from awscli.customizations.s3.utils import ProvideUploadContentTypeSubscriber
-from awscli.customizations.s3.utils import ProvideCopyContentTypeSubscriber
-from awscli.customizations.s3.utils import ProvideLastModifiedTimeSubscriber
-from awscli.customizations.s3.utils import DirectoryCreatorSubscriber
-from awscli.customizations.s3.utils import DeleteSourceFileSubscriber
-from awscli.customizations.s3.utils import DeleteSourceObjectSubscriber
-from awscli.customizations.s3.utils import DeleteCopySourceObjectSubscriber
 from awscli.compat import get_binary_stdin
-
+from awscli.customizations.s3.results import (
+    CommandResultRecorder,
+    DoneResultSubscriber,
+    DryRunResult,
+    FailureResult,
+    NoProgressResultPrinter,
+    OnlyShowErrorsResultPrinter,
+    ProgressResultSubscriber,
+    QueuedResult,
+    QueuedResultSubscriber,
+    ResultPrinter,
+    ResultProcessor,
+    ResultRecorder,
+    SuccessResult,
+)
+from awscli.customizations.s3.subscribers import (
+    CopyPropsSubscriberFactory,
+    DeleteCopySourceObjectSubscriber,
+    DeleteSourceFileSubscriber,
+    DeleteSourceObjectSubscriber,
+    DirectoryCreatorSubscriber,
+    ProvideETagSubscriber,
+    ProvideLastModifiedTimeSubscriber,
+    ProvideSizeSubscriber,
+    ProvideUploadContentTypeSubscriber,
+)
+from awscli.customizations.s3.transferconfig import (
+    create_transfer_config_from_runtime_config,
+)
+from awscli.customizations.s3.utils import (
+    MAX_UPLOAD_SIZE,
+    NonSeekableStream,
+    RequestParamsMapper,
+    StdoutBytesWriter,
+    create_warning,
+    find_bucket_key,
+    human_readable_size,
+    relative_path,
+)
 
 LOGGER = logging.getLogger(__name__)
 
 
-class S3TransferHandlerFactory(object):
-    MAX_IN_MEMORY_CHUNKS = 6
-
-    def __init__(self, cli_params, runtime_config):
+class S3TransferHandlerFactory:
+    def __init__(self, cli_params):
         """Factory for S3TransferHandlers
 
         :type cli_params: dict
         :param cli_params: The parameters provide to the CLI command
-
-        :type runtime_config: RuntimeConfig
-        :param runtime_config: The runtime config for the CLI command
-            being run
         """
         self._cli_params = cli_params
-        self._runtime_config = runtime_config
 
-    def __call__(self, client, result_queue):
+    def __call__(self, transfer_manager, result_queue):
         """Creates a S3TransferHandler instance
 
-        :type client: botocore.client.Client
-        :param client: The client to power the S3TransferHandler
+        :param transfer_manager: The transfer manager to power the
+            S3TransferHandler
 
         :type result_queue: queue.Queue
         :param result_queue: The result queue to be used to process results
@@ -81,29 +80,19 @@ class S3TransferHandlerFactory(object):
 
         :returns: A S3TransferHandler instance
         """
-        transfer_config = create_transfer_config_from_runtime_config(
-            self._runtime_config)
-        transfer_config.max_in_memory_upload_chunks = self.MAX_IN_MEMORY_CHUNKS
-        transfer_config.max_in_memory_download_chunks = \
-            self.MAX_IN_MEMORY_CHUNKS
-
-        transfer_manager = TransferManager(client, transfer_config)
-
-        LOGGER.debug(
-            "Using a multipart threshold of %s and a part size of %s",
-            transfer_config.multipart_threshold,
-            transfer_config.multipart_chunksize
-        )
         result_recorder = ResultRecorder()
         result_processor_handlers = [result_recorder]
         self._add_result_printer(result_recorder, result_processor_handlers)
         result_processor = ResultProcessor(
-            result_queue, result_processor_handlers)
+            result_queue, result_processor_handlers
+        )
         command_result_recorder = CommandResultRecorder(
-            result_queue, result_recorder, result_processor)
+            result_queue, result_recorder, result_processor
+        )
 
         return S3TransferHandler(
-            transfer_manager, self._cli_params, command_result_recorder)
+            transfer_manager, self._cli_params, command_result_recorder
+        )
 
     def _add_result_printer(self, result_recorder, result_processor_handlers):
         if self._cli_params.get('quiet'):
@@ -115,11 +104,15 @@ class S3TransferHandlerFactory(object):
         elif not self._cli_params.get('progress'):
             result_printer = NoProgressResultPrinter(result_recorder)
         else:
-            result_printer = ResultPrinter(result_recorder)
+            result_printer = ResultPrinter(
+                result_recorder,
+                frequency=self._cli_params.get('progress_frequency'),
+                oneline=not self._cli_params.get('progress_multiline'),
+            )
         result_processor_handlers.append(result_printer)
 
 
-class S3TransferHandler(object):
+class S3TransferHandler:
     def __init__(self, transfer_manager, cli_params, result_command_recorder):
         """Backend for performing S3 transfers
 
@@ -144,8 +137,9 @@ class S3TransferHandler(object):
         self._result_command_recorder = result_command_recorder
 
         submitter_args = (
-            self._transfer_manager, self._result_command_recorder.result_queue,
-            cli_params
+            self._transfer_manager,
+            self._result_command_recorder.result_queue,
+            cli_params,
         )
         self._submitters = [
             UploadStreamRequestSubmitter(*submitter_args),
@@ -154,7 +148,7 @@ class S3TransferHandler(object):
             DownloadRequestSubmitter(*submitter_args),
             CopyRequestSubmitter(*submitter_args),
             DeleteRequestSubmitter(*submitter_args),
-            LocalDeleteRequestSubmitter(*submitter_args)
+            LocalDeleteRequestSubmitter(*submitter_args),
         ]
 
     def call(self, fileinfos):
@@ -178,13 +172,13 @@ class S3TransferHandler(object):
                                 total_submissions += 1
                             break
                 self._result_command_recorder.notify_total_submissions(
-                    total_submissions)
+                    total_submissions
+                )
         return self._result_command_recorder.get_command_result()
 
 
-class BaseTransferRequestSubmitter(object):
+class BaseTransferRequestSubmitter:
     REQUEST_MAPPER_METHOD = None
-    RESULT_SUBSCRIBER_CLASS = None
 
     def __init__(self, transfer_manager, result_queue, cli_params):
         """Submits transfer requests to the TransferManager
@@ -243,30 +237,47 @@ class BaseTransferRequestSubmitter(object):
         extra_args = {}
         if self.REQUEST_MAPPER_METHOD:
             self.REQUEST_MAPPER_METHOD(extra_args, self._cli_params)
-        subscribers = []
-        self._add_additional_subscribers(subscribers, fileinfo)
-        # The result subscriber class should always be the last registered
-        # subscriber to ensure it is not missing any information that
-        # may have been added in a different subscriber such as size.
-        if self.RESULT_SUBSCRIBER_CLASS:
-            result_kwargs = {'result_queue': self._result_queue}
-            if self._cli_params.get('is_move', False):
-                result_kwargs['transfer_type'] = 'move'
-            subscribers.append(self.RESULT_SUBSCRIBER_CLASS(**result_kwargs))
-
         if not self._cli_params.get('dryrun'):
             return self._submit_transfer_request(
-                fileinfo, extra_args, subscribers)
+                fileinfo, extra_args, self._get_subscribers(fileinfo)
+            )
         else:
             self._submit_dryrun(fileinfo)
 
-    def _submit_dryrun(self, fileinfo):
-        transfer_type = fileinfo.operation_name
-        if self._cli_params.get('is_move', False):
-            transfer_type = 'move'
+    def _get_subscribers(self, fileinfo):
+        subscribers = []
+        result_subscriber_kwargs = self._get_result_subscriber_kwargs(fileinfo)
+        self._add_provide_size_subscriber(subscribers, fileinfo)
+        subscribers.append(ProvideETagSubscriber(fileinfo.etag))
+        subscribers.append(QueuedResultSubscriber(**result_subscriber_kwargs))
+        self._add_additional_subscribers(subscribers, fileinfo)
+        subscribers.extend(
+            [
+                ProgressResultSubscriber(**result_subscriber_kwargs),
+                DoneResultSubscriber(**result_subscriber_kwargs),
+            ]
+        )
+        return subscribers
+
+    def _get_result_subscriber_kwargs(self, fileinfo):
         src, dest = self._format_src_dest(fileinfo)
-        self._result_queue.put(DryRunResult(
-            transfer_type=transfer_type, src=src, dest=dest))
+        return {
+            'result_queue': self._result_queue,
+            'transfer_type': self._get_transfer_type(fileinfo),
+            'src': src,
+            'dest': dest,
+            'size': fileinfo.size,
+        }
+
+    def _submit_dryrun(self, fileinfo):
+        transfer_type = self._get_transfer_type(fileinfo)
+        src, dest = self._format_src_dest(fileinfo)
+        self._result_queue.put(
+            DryRunResult(transfer_type=transfer_type, src=src, dest=dest)
+        )
+
+    def _add_provide_size_subscriber(self, subscribers, fileinfo):
+        subscribers.append(ProvideSizeSubscriber(fileinfo.size))
 
     def _add_additional_subscribers(self, subscribers, fileinfo):
         pass
@@ -278,7 +289,7 @@ class BaseTransferRequestSubmitter(object):
         for warning_handler in self._get_warning_handlers():
             if warning_handler(fileinfo):
                 # On the first warning handler that returns a signal to skip
-                # immediately propagate this signal and no longer check
+                # immediately propogate this signal and no longer check
                 # the other warning handlers as no matter what the file will
                 # be skipped.
                 return True
@@ -291,27 +302,27 @@ class BaseTransferRequestSubmitter(object):
         return []
 
     def _should_inject_content_type(self):
-        return (
-            self._cli_params.get('guess_mime_type') and
-            not self._cli_params.get('content_type')
-        )
+        return self._cli_params.get(
+            'guess_mime_type'
+        ) and not self._cli_params.get('content_type')
 
     def _warn_glacier(self, fileinfo):
         if not self._cli_params.get('force_glacier_transfer'):
             if not fileinfo.is_glacier_compatible():
                 LOGGER.debug(
                     'Encountered glacier object s3://%s. Not performing '
-                    '%s on object.' % (fileinfo.src, fileinfo.operation_name))
+                    '%s on object.' % (fileinfo.src, fileinfo.operation_name)
+                )
                 if not self._cli_params.get('ignore_glacier_warnings'):
                     warning = create_warning(
-                        's3://'+fileinfo.src,
+                        's3://' + fileinfo.src,
                         'Object is of storage class GLACIER. Unable to '
                         'perform %s operations on GLACIER objects. You must '
                         'restore the object to be able to perform the '
                         'operation. See aws s3 %s help for additional '
                         'parameter options to ignore or force these '
-                        'transfers.' %
-                        (fileinfo.operation_name, fileinfo.operation_name)
+                        'transfers.'
+                        % (fileinfo.operation_name, fileinfo.operation_name),
                     )
                     self._result_queue.put(warning)
                 return True
@@ -322,14 +333,21 @@ class BaseTransferRequestSubmitter(object):
         # need to take that into account when checking for a parent prefix.
         parent_prefix = '..' + os.path.sep
         escapes_cwd = os.path.normpath(fileinfo.compare_key).startswith(
-            parent_prefix)
+            parent_prefix
+        )
         if escapes_cwd:
             warning = create_warning(
-                fileinfo.compare_key, "File references a parent directory.")
+                fileinfo.compare_key, "File references a parent directory."
+            )
             self._result_queue.put(warning)
             return True
         return False
 
+    def _get_transfer_type(self, fileinfo):
+        if self._cli_params.get('is_move', False):
+            return 'move'
+        return fileinfo.operation_name
+
     def _format_src_dest(self, fileinfo):
         """Returns formatted versions of a fileinfos source and destination."""
         raise NotImplementedError('_format_src_dest')
@@ -345,13 +363,11 @@ class BaseTransferRequestSubmitter(object):
 
 class UploadRequestSubmitter(BaseTransferRequestSubmitter):
     REQUEST_MAPPER_METHOD = RequestParamsMapper.map_put_object_params
-    RESULT_SUBSCRIBER_CLASS = UploadResultSubscriber
 
     def can_submit(self, fileinfo):
         return fileinfo.operation_name == 'upload'
 
     def _add_additional_subscribers(self, subscribers, fileinfo):
-        subscribers.append(ProvideSizeSubscriber(fileinfo.size))
         if self._should_inject_content_type():
             subscribers.append(ProvideUploadContentTypeSubscriber())
         if self._cli_params.get('is_move', False):
@@ -361,8 +377,11 @@ class UploadRequestSubmitter(BaseTransferRequestSubmitter):
         bucket, key = find_bucket_key(fileinfo.dest)
         filein = self._get_filein(fileinfo)
         return self._transfer_manager.upload(
-            fileobj=filein, bucket=bucket, key=key,
-            extra_args=extra_args, subscribers=subscribers
+            fileobj=filein,
+            bucket=bucket,
+            key=key,
+            extra_args=extra_args,
+            subscribers=subscribers,
         )
 
     def _get_filein(self, fileinfo):
@@ -374,11 +393,13 @@ class UploadRequestSubmitter(BaseTransferRequestSubmitter):
     def _warn_if_too_large(self, fileinfo):
         if getattr(fileinfo, 'size') and fileinfo.size > MAX_UPLOAD_SIZE:
             file_path = relative_path(fileinfo.src)
-            warning_message = (
-                "File %s exceeds s3 upload limit of %s." % (
-                    file_path, human_readable_size(MAX_UPLOAD_SIZE)))
+            warning_message = "File %s exceeds s3 upload limit of %s." % (
+                file_path,
+                human_readable_size(MAX_UPLOAD_SIZE),
+            )
             warning = create_warning(
-                file_path, warning_message, skip_file=False)
+                file_path, warning_message, skip_file=False
+            )
             self._result_queue.put(warning)
 
     def _format_src_dest(self, fileinfo):
@@ -389,27 +410,31 @@ class UploadRequestSubmitter(BaseTransferRequestSubmitter):
 
 class DownloadRequestSubmitter(BaseTransferRequestSubmitter):
     REQUEST_MAPPER_METHOD = RequestParamsMapper.map_get_object_params
-    RESULT_SUBSCRIBER_CLASS = DownloadResultSubscriber
 
     def can_submit(self, fileinfo):
         return fileinfo.operation_name == 'download'
 
     def _add_additional_subscribers(self, subscribers, fileinfo):
-        subscribers.append(ProvideSizeSubscriber(fileinfo.size))
-        subscribers.append(ProvideETagSubscriber(fileinfo.etag))
         subscribers.append(DirectoryCreatorSubscriber())
-        subscribers.append(ProvideLastModifiedTimeSubscriber(
-            fileinfo.last_update, self._result_queue))
+        subscribers.append(
+            ProvideLastModifiedTimeSubscriber(
+                fileinfo.last_update, self._result_queue
+            )
+        )
         if self._cli_params.get('is_move', False):
-            subscribers.append(DeleteSourceObjectSubscriber(
-                fileinfo.source_client))
+            subscribers.append(
+                DeleteSourceObjectSubscriber(fileinfo.source_client)
+            )
 
     def _submit_transfer_request(self, fileinfo, extra_args, subscribers):
         bucket, key = find_bucket_key(fileinfo.src)
         fileout = self._get_fileout(fileinfo)
         return self._transfer_manager.download(
-            fileobj=fileout, bucket=bucket, key=key,
-            extra_args=extra_args, subscribers=subscribers
+            fileobj=fileout,
+            bucket=bucket,
+            key=key,
+            extra_args=extra_args,
+            subscribers=subscribers,
         )
 
     def _get_fileout(self, fileinfo):
@@ -426,28 +451,37 @@ class DownloadRequestSubmitter(BaseTransferRequestSubmitter):
 
 class CopyRequestSubmitter(BaseTransferRequestSubmitter):
     REQUEST_MAPPER_METHOD = RequestParamsMapper.map_copy_object_params
-    RESULT_SUBSCRIBER_CLASS = CopyResultSubscriber
 
     def can_submit(self, fileinfo):
         return fileinfo.operation_name == 'copy'
 
     def _add_additional_subscribers(self, subscribers, fileinfo):
-        subscribers.append(ProvideSizeSubscriber(fileinfo.size))
-        subscribers.append(ProvideETagSubscriber(fileinfo.etag))
-        if self._should_inject_content_type():
-            subscribers.append(ProvideCopyContentTypeSubscriber())
+        if not self._cli_params.get('metadata_directive'):
+            self._add_copy_props_subscribers(subscribers, fileinfo)
         if self._cli_params.get('is_move', False):
-            subscribers.append(DeleteCopySourceObjectSubscriber(
-                fileinfo.source_client))
+            subscribers.append(
+                DeleteCopySourceObjectSubscriber(fileinfo.source_client)
+            )
+
+    def _add_copy_props_subscribers(self, subscribers, fileinfo):
+        copy_props_factory = CopyPropsSubscriberFactory(
+            self._transfer_manager.client,
+            self._transfer_manager.config,
+            self._cli_params,
+        )
+        subscribers.extend(copy_props_factory.get_subscribers(fileinfo))
 
     def _submit_transfer_request(self, fileinfo, extra_args, subscribers):
         bucket, key = find_bucket_key(fileinfo.dest)
         source_bucket, source_key = find_bucket_key(fileinfo.src)
         copy_source = {'Bucket': source_bucket, 'Key': source_key}
         return self._transfer_manager.copy(
-            bucket=bucket, key=key, copy_source=copy_source,
-            extra_args=extra_args, subscribers=subscribers,
-            source_client=fileinfo.source_client
+            bucket=bucket,
+            key=key,
+            copy_source=copy_source,
+            extra_args=extra_args,
+            subscribers=subscribers,
+            source_client=fileinfo.source_client,
         )
 
     def _get_warning_handlers(self):
@@ -460,19 +494,19 @@ class CopyRequestSubmitter(BaseTransferRequestSubmitter):
 
 
 class UploadStreamRequestSubmitter(UploadRequestSubmitter):
-    RESULT_SUBSCRIBER_CLASS = UploadStreamResultSubscriber
-
     def can_submit(self, fileinfo):
-        return (
-            fileinfo.operation_name == 'upload' and
-            self._cli_params.get('is_stream')
+        return fileinfo.operation_name == 'upload' and self._cli_params.get(
+            'is_stream'
         )
 
-    def _add_additional_subscribers(self, subscribers, fileinfo):
+    def _add_provide_size_subscriber(self, subscribers, fileinfo):
         expected_size = self._cli_params.get('expected_size', None)
         if expected_size is not None:
             subscribers.append(ProvideSizeSubscriber(int(expected_size)))
 
+    def _add_additional_subscribers(self, subscribers, fileinfo):
+        pass
+
     def _get_filein(self, fileinfo):
         binary_stdin = get_binary_stdin()
         return NonSeekableStream(binary_stdin)
@@ -482,14 +516,14 @@ class UploadStreamRequestSubmitter(UploadRequestSubmitter):
 
 
 class DownloadStreamRequestSubmitter(DownloadRequestSubmitter):
-    RESULT_SUBSCRIBER_CLASS = DownloadStreamResultSubscriber
-
     def can_submit(self, fileinfo):
-        return (
-            fileinfo.operation_name == 'download' and
-            self._cli_params.get('is_stream')
+        return fileinfo.operation_name == 'download' and self._cli_params.get(
+            'is_stream'
         )
 
+    def _add_provide_size_subscriber(self, subscribers, fileinfo):
+        pass
+
     def _add_additional_subscribers(self, subscribers, fileinfo):
         pass
 
@@ -502,17 +536,23 @@ class DownloadStreamRequestSubmitter(DownloadRequestSubmitter):
 
 class DeleteRequestSubmitter(BaseTransferRequestSubmitter):
     REQUEST_MAPPER_METHOD = RequestParamsMapper.map_delete_object_params
-    RESULT_SUBSCRIBER_CLASS = DeleteResultSubscriber
 
     def can_submit(self, fileinfo):
-        return fileinfo.operation_name == 'delete' and \
-            fileinfo.src_type == 's3'
+        return (
+            fileinfo.operation_name == 'delete' and fileinfo.src_type == 's3'
+        )
+
+    def _add_provide_size_subscriber(self, subscribers, fileinfo):
+        pass
 
     def _submit_transfer_request(self, fileinfo, extra_args, subscribers):
         bucket, key = find_bucket_key(fileinfo.src)
         return self._transfer_manager.delete(
-            bucket=bucket, key=key, extra_args=extra_args,
-            subscribers=subscribers)
+            bucket=bucket,
+            key=key,
+            extra_args=extra_args,
+            subscribers=subscribers,
+        )
 
     def _format_src_dest(self, fileinfo):
         return self._format_s3_path(fileinfo.src), None
@@ -520,11 +560,12 @@ class DeleteRequestSubmitter(BaseTransferRequestSubmitter):
 
 class LocalDeleteRequestSubmitter(BaseTransferRequestSubmitter):
     REQUEST_MAPPER_METHOD = None
-    RESULT_SUBSCRIBER_CLASS = None
 
     def can_submit(self, fileinfo):
-        return fileinfo.operation_name == 'delete' and \
-            fileinfo.src_type == 'local'
+        return (
+            fileinfo.operation_name == 'delete'
+            and fileinfo.src_type == 'local'
+        )
 
     def _submit_transfer_request(self, fileinfo, extra_args, subscribers):
         # This is quirky but essentially instead of relying on a built-in
@@ -540,19 +581,15 @@ class LocalDeleteRequestSubmitter(BaseTransferRequestSubmitter):
         # deleting a local file only happens for sync --delete downloads and
         # is very fast compared to all of the other types of transfers.
         src, dest = self._format_src_dest(fileinfo)
-        result_kwargs = {
-            'transfer_type': 'delete',
-            'src': src,
-            'dest': dest
-        }
+        result_kwargs = {'transfer_type': 'delete', 'src': src, 'dest': dest}
         try:
-            self._result_queue.put(QueuedResult(
-                total_transfer_size=0, **result_kwargs))
+            self._result_queue.put(
+                QueuedResult(total_transfer_size=0, **result_kwargs)
+            )
             os.remove(fileinfo.src)
             self._result_queue.put(SuccessResult(**result_kwargs))
         except Exception as e:
-            self._result_queue.put(
-                FailureResult(exception=e, **result_kwargs))
+            self._result_queue.put(FailureResult(exception=e, **result_kwargs))
         finally:
             # Return True to indicate that the transfer was submitted
             return True
